/***********************************************************************
 *
 * Name:        Scheduler.c
 *
 * Description: Time controlling module for Video Plugin               
 *
 * Author:	Tomasz Stachowiak, NPAC
 *
 * Date:	November 1996
 *
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#include "VideoProtocol.h"

#define MAX_USERS 20

#define DELAY_MAX 200
#define DELAY_MIN 50
#define RATE_STEP 1
#define SPACE_MIN 5
#define INTERVAL_STEP 50
#define SPEED_RATIO 0.1

typedef struct timeval Time;

typedef struct _usrTimeInfo
{
  int userNo;                    /* conference member ID */
  int rate;
  Time startT;                   /* time of receiving first frame */ 
  Time startStamp;               /* first frame time stamp */
  Time lastStamp;                /* previous frame time stamp */
  Time lastFrameT;               /* time of receiving first frame */
  int framesReceived;
  Time totalT;
  int framesLate;
  int framesAvanced;
  int jitter;
  int totalBytes;
} UsrTimeInfo;

typedef struct _sendTimeInfo
{
  float rate;
  int interval;
  Time startT;
  Time lastFrameT;
  Time nextFrameT;
  int framesSent;
  Time totalT;
  int slowNo;
  int speedNo;
  int totalBytes;
} SendTimeInfo;

void slowDown();
void speedUp();
void timeSum(Time *a, int b, Time *result);
int timeDiff(Time *a, Time *b);

UsrTimeInfo usrTimeInfo[MAX_USERS];
SendTimeInfo sendTimeInfo;
static int rec_first[MAX_USERS];
static int send_first = 1;
int max_rate = 0;
int canChange = 1;

extern float delayCoeff;
extern state;

/*********************************************************************
 * Procedure handles received frame, it does all the statistics and
 * optionally it changes the sending rate 
 *
 *********************************************************************/
void doReceiveStatistics(int userNo, Time *frameTimeStamp)
{
  Time currentT;
  int diff, local_diff, remote_diff;

  gettimeofday(&currentT);
  if(!rec_first[userNo])
    {
      local_diff = timeDiff(&currentT, &usrTimeInfo[userNo].startT);
      remote_diff = timeDiff(frameTimeStamp, &usrTimeInfo[userNo].startStamp);
      diff = local_diff - remote_diff;
      printf("Delay %d ms\n", diff);
      if(diff > DELAY_MAX)
	{
	  delayCoeff = 2;
	  usrTimeInfo[userNo].framesLate++;
	}
      if(diff < DELAY_MIN)
	{
	  delayCoeff = 1.1;
	}
    }
  else
    {
      gettimeofday(&usrTimeInfo[userNo].startT);
      usrTimeInfo[userNo].startStamp.tv_sec = frameTimeStamp->tv_sec;
      usrTimeInfo[userNo].startStamp.tv_usec = frameTimeStamp->tv_usec;
      rec_first[userNo] = 0;
    }
  usrTimeInfo[userNo].lastStamp.tv_sec = frameTimeStamp->tv_sec;
  usrTimeInfo[userNo].lastStamp.tv_usec = frameTimeStamp->tv_usec;
  usrTimeInfo[userNo].framesReceived++;
  usrTimeInfo[userNo].lastFrameT.tv_sec = currentT.tv_sec;
  usrTimeInfo[userNo].lastFrameT.tv_usec = currentT.tv_usec;
  /* ...calculate jitter */
}

void receiveStart(int userNo)
{
  gettimeofday(&usrTimeInfo[userNo].startT);
}

void receiveStop(int userNo)
{
  Time tmp;

  gettimeofday(&tmp);
  timeSum(&usrTimeInfo[userNo].totalT, 
	  timeDiff(&tmp, &usrTimeInfo[userNo].startT), 
	  &usrTimeInfo[userNo].totalT);
  rec_first[userNo] = 1;
}

void receiveTimeInit()
{
  int i;
  
  for(i=0; i<MAX_USERS; i++)
    {
      usrTimeInfo[i].userNo = -1;
      rec_first[i] = 1;
    }
}

void usrTimeInit(int userNo)
{
  usrTimeInfo[userNo].userNo = userNo;
  gettimeofday(&usrTimeInfo[userNo].startT);
  usrTimeInfo[userNo].framesReceived = 0;
  usrTimeInfo[userNo].totalT.tv_sec = 0;
  usrTimeInfo[userNo].totalT.tv_usec = 0;
  usrTimeInfo[userNo].framesLate = 0;
  usrTimeInfo[userNo].framesAvanced = 0;
  usrTimeInfo[userNo].jitter = 0;
  usrTimeInfo[userNo].totalBytes = 0;
  
}

void usrTimeReport()
{
  int i;

  for(i=0; i<MAX_USERS; i++)
    if(usrTimeInfo[i].userNo >= 0)
      {
	printf("----------------------------------------------\n");
	printf("USER: %d\n", i);
	printf("----------------------------------------------\n");
	printf("Frames received: %d\n",usrTimeInfo[i].framesReceived);
	printf("Total time:      %d\n", 
	       (usrTimeInfo[i].totalT.tv_sec*1000)
	       +(usrTimeInfo[i].totalT.tv_usec/1000));
	printf("Frames late:     %d\n", usrTimeInfo[i].framesLate);
	printf("Frames avanced:  %d\n", usrTimeInfo[i].framesAvanced);
	printf("----------------------------------------------\n\n");
	fflush(stdout);
      }
}

/*********************************************************************
 * Procedure schedules next frame to send based on the sendTimeInfo
 * It return the time(in miliseconds) left to next frame sending 
 *
 *********************************************************************/
int doSendSchedule()
{
  Time currentT;
  int diff, togo;
  static int tmp = 0;
  float rate;
  
  printf("Sending rate %f\n",(float)1000/(float)sendTimeInfo.interval); 
  canChange = 1;
  if(!send_first)
    {
      gettimeofday(&currentT);
      diff = timeDiff(&currentT, &sendTimeInfo.lastFrameT);
      togo = sendTimeInfo.interval - diff;
      printf("Wait %d ms\n", togo);
      if(togo < (float)sendTimeInfo.interval*0.1)
	{
	  max_rate = 1;
	  slowDown();
	}
      else
	{
	  tmp = (tmp + 1) % 1;
	  if(tmp == 0)
	    max_rate = 0;
	}
      if(togo <= 0)
	{
	  return 0;
	}
      else
	{
	  return togo;
	}
    }
  else
    {
      send_first = 0;
      return 0;
    }
}

void doSendStatistics(int bytes)
{
  Time currentT;

  gettimeofday(&currentT);
  sendTimeInfo.rate = (float)1000 / (float)timeDiff(&currentT, &sendTimeInfo.lastFrameT);
  /*  printf("Sending rate: %f\n", sendTimeInfo.rate); */
  sendTimeInfo.lastFrameT.tv_sec = currentT.tv_sec;
  sendTimeInfo.lastFrameT.tv_usec = currentT.tv_usec;
  sendTimeInfo.framesSent++;
  sendTimeInfo.totalBytes += bytes;
}



void slowDown()
{
      sendTimeInfo.interval = (int)((float)sendTimeInfo.interval 
				    * ((float)1+(float)SPEED_RATIO)); 
      /*      if(sendTimeInfo.interval > 4000)
	sendTimeInfo.interval = 4000;*/
      sendTimeInfo.slowNo++;
      canChange = 0;
}

void speedUp()
{
  if(canChange)
    {
      if((state == PLAY) && (!max_rate))
	{
	  sendTimeInfo.interval = (int)((float)sendTimeInfo.interval 
					* ((float)1-(float)SPEED_RATIO)); 
	  
	  if(sendTimeInfo.interval < 200)
	    sendTimeInfo.interval = 200;
	  if(sendTimeInfo.interval > 4000)
	    sendTimeInfo.interval = 4000;
	  sendTimeInfo.speedNo++;
	}
      canChange = 0;
    }
}

void sendStart()
{
  gettimeofday(&sendTimeInfo.startT);
  sendTimeInfo.interval = 200;
}

void sendStop()
{
  Time tmp;

  gettimeofday(&tmp);
  timeSum(&sendTimeInfo.totalT, 
	  timeDiff(&tmp, &sendTimeInfo.startT), 
	  &sendTimeInfo.totalT);
  send_first = 1;
}

void sendTimeInit()
{
  sendTimeInfo.rate = 5;
  sendTimeInfo.interval = 200;
  gettimeofday(&sendTimeInfo.startT);
  sendTimeInfo.framesSent = 0;
  sendTimeInfo.totalT.tv_sec = 0;
  sendTimeInfo.totalT.tv_usec = 0;
  sendTimeInfo.slowNo = 0;
  sendTimeInfo.speedNo = 0;
  sendTimeInfo.totalBytes = 0;
  
}

void sendTimeReport()
{
  char msg[500];
  float frameRate, bytesRate, bytesFrame;

  frameRate = (float)sendTimeInfo.framesSent/(float)sendTimeInfo.totalT.tv_sec;
  bytesRate = (float)sendTimeInfo.totalBytes/(float)sendTimeInfo.totalT.tv_sec;
  bytesFrame = (float)sendTimeInfo.totalBytes/(float)sendTimeInfo.framesSent;
  sprintf(msg, "        STATISTICS\n--------------------------\nFrames sent:   %d\nTotal time:    %d ms\nBytes/frame:   %f\nBytes/second:  %f\nFrames/second: %f\n", 
	  sendTimeInfo.framesSent, 
	  (sendTimeInfo.totalT.tv_sec*1000)+(sendTimeInfo.totalT.tv_usec/1000),
	  bytesFrame, 
	  bytesRate, 
	  frameRate);
  showMessage(msg);
}

/********************************************************************
 * 
 * Procedure operating on timeval structures
 *
 *******************************************************************/
/* Sum result = a + b (b in miliseconds)*/
void timeSum(Time *a, int b, Time *result)
{
  int pom;

  result->tv_sec = a->tv_sec + (int)(b / 1000);
  result->tv_usec = (b % 1000) * 1000;
}

/* Difference a-b in miliseconds */
int timeDiff(Time *a, Time *b)
{
  int result;

  result = 1000 * (a->tv_sec - b->tv_sec) + (a->tv_usec - b->tv_usec)/1000;

  return result;
}



















